#define CORE_PRIVATE

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_core.h"	/* manipulate core_server_config in set_document_root */

#include "ap_config.h"
#include "apr_strings.h"
#include "apr_tables.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <cdb.h>

#if !defined(WIN32) && !defined(OS2) && !defined(BEOS) && !defined(NETWARE)
#define HAVE_UNIX_SUEXEC
#endif

#ifdef HAVE_UNIX_SUEXEC
#include "unixd.h"
#endif

module AP_MODULE_DECLARE_DATA vhost_cdb_module;

typedef struct {
	char			enable;
	char 			*filename;
	char			*defaulthost;
	apr_array_header_t	*export_to_env;
	apr_array_header_t	*export_to_notes;
	
} mod_vhost_cdb_config;

/************************************************************************************************/
/* function prototypes										*/
/************************************************************************************************/

static char 		*db_get(request_rec *, struct cdb *, const char *);

static char 		*get_config_from_db(request_rec *, struct cdb *, const char *, mod_vhost_cdb_config *);
static apr_table_t 	*create_config_table(request_rec *, const char *);
static int 		mod_vhost_cdb_set_document_root(request_rec *, apr_table_t *);
int 			mod_vhost_cdb_set_script_alias(request_rec *, apr_table_t *);
int 			mod_vhost_cdb_export_vars(request_rec *, apr_table_t *);
static int 		expand_path(request_rec *, apr_table_t *, const char *);

static int 		mod_vhost_cdb_trans_uri(request_rec *);

#ifdef HAVE_UNIX_SUEXEC
static ap_unix_identity_t *mod_vhost_cdb_suexec(const request_rec *);
#endif


/************************************************************************************************/
/* db_get(request_rec, config, hostname)							*/
/* returns the value (a query-string-like char *) which fits to hostname			*/
/************************************************************************************************/

static char *db_get(request_rec *r, struct cdb *db, const char *hostname)
{
/* ap_log_error(APLOG_MARK,APLOG_DEBUG,0,r->server,"db_get: trying to find %s",hostname); */

if(!cdb_find(db,hostname,strlen(hostname)))
	{
	return NULL;
	}

unsigned int pos=cdb_datapos(db);
unsigned int len=cdb_datalen(db);
char *entry=NULL;

if(len>0)
	{
	entry=apr_pcalloc(r->pool,sizeof(char)*(len+1));
	if(cdb_read(db,entry,len,pos)<0)
		{
		ap_log_error(APLOG_MARK,APLOG_CRIT,0,r->server,"db_get: cdb_read failed: %s",strerror(errno));
		return NULL;
		}
	}

return entry;
}

/************************************************************************************************/
/* apr_table_t *create_config_table(request_rec *, char *db_result)				*/
/* creates a table from a tab-separated "key=value" char *					*/
/* sample: "foo=bar	baz=mjoelk"								*/
/************************************************************************************************/

static apr_table_t *create_config_table(request_rec *r, const char *db_result)
{
char *qstr=(char *)db_result;

apr_table_t *rec=apr_table_make(r->pool,4);

char *next_token=NULL;
char *sep=NULL;

do
	{
	next_token=index(qstr,'\t');
	sep=index(qstr,'=');

	if(next_token==NULL)
		{
		next_token=qstr+strlen(qstr); // *next_token==\0
		}

	char *val=NULL;

	if(sep==NULL || (sep>next_token))
		{
		val=apr_pcalloc(r->pool,1);
		sep=next_token;
		}
	else
		{
		val=apr_pstrmemdup(r->pool,sep+1,next_token-sep-1);
		}

	char *key=apr_pstrmemdup(r->pool,qstr,sep-qstr);

	/* 
	 * we now have key and value parsed
	 * lets bring that into the db_rec
	 */	

	apr_table_addn(rec,key,val);

	/* ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"create_config_table: parsed %s(%s)",key,val); */
	
	if(*next_token!=0) qstr=next_token+1;

	}
while(*next_token!=0);

return rec;
}

/* ****************************************************************************	*/
/* mod_vhost_cdb_set_document_root						*/
/* sets all needed variables to apply the DocumentRoot from cdb			*/
/* ****************************************************************************	*/

static int mod_vhost_cdb_set_document_root(request_rec *r, apr_table_t *conf)
{
if(r->hostname!=NULL)
	{
	r->server->server_hostname = apr_pstrdup(r->server->process->pool, r->hostname);
	r->parsed_uri.hostinfo = apr_pstrdup(r->pool,r->hostname);
	r->parsed_uri.hostname = apr_pstrdup(r->pool,r->hostname);
	}

r->server->is_virtual = 1;

/* ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"set_doc_root: server_hostname: %s",r->server->server_hostname); */

if(!mod_vhost_cdb_set_script_alias(r,conf))
	{
	if(r->uri[0]!='/')
		{
		return 0;
		}
	r->filename=apr_pstrcat(r->pool,apr_table_get(conf,"DocumentRoot"),r->uri,NULL);
	ap_no2slash(r->filename);
	}

char filter[1024];
snprintf(filter,1024,"%s",apr_table_get(conf,"DocumentRoot"));
ap_no2slash(filter);
apr_table_setn(r->subprocess_env, "SERVER_ROOT", apr_pstrdup(r->pool,filter));
apr_table_setn(r->subprocess_env, "DOCUMENT_ROOT", apr_pstrdup(r->pool,filter));

/*
 * put the document_root into the core configuration
 * otherwise ENV{'DOCUMENT_ROOT'} will be overwritten later
 */

core_server_config *core=(core_server_config *) ap_get_module_config(r->server->module_config, &core_module);
core->ap_document_root = apr_pstrdup(r->server->process->pool, apr_table_get(conf,"DocumentRoot"));

return 1;
}


/* ****************************************************************************	*/
/* mod_vhost_cdb_set_script_alias						*/
/* sets all needed variables to apply the ScriptAlias from cdb			*/
/* returns 1 if the request was a cgi-bin/-request				*/
/*										*/
/* code is actually borrowed from mod_vhost_ldap.c				*/
/* ****************************************************************************	*/

int mod_vhost_cdb_set_script_alias(request_rec *r, apr_table_t *conf)
{
const char *script_alias=apr_table_get(conf,"ScriptAlias");
if(script_alias==NULL)
	{
	ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"mod_vhost_cdb_set_script_alias: no configured ScriptAlias");
	return 0;
	}

char *cgi=strstr(r->uri, "cgi-bin/");
if (!cgi || (cgi && (cgi!=r->uri + strspn(r->uri, "/"))) )
	{
	return 0;
	}

/* ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"mod_vhost_cdb_set_script_alias: setting script_alias, %s%s",script_alias,cgi+strlen("cgi-bin")); */

r->filename=apr_pstrcat(r->pool, script_alias, cgi + strlen("cgi-bin"), NULL);
r->handler=apr_pstrcat(r->pool,"cgi-script",NULL);
apr_table_setn(r->notes, "alias-forced-type", r->handler);

return 1;
}

/* ****************************************************************************	*/
/* mod_vhost_cdb_export_vars							*/
/* exports variables from cdb-entry to subprocess_env/notes when module		*/
/* is configured to do so via VhostCDB_Export*					*/
/* ****************************************************************************	*/

int mod_vhost_cdb_export_vars(request_rec *r, apr_table_t *db_conf)
{
mod_vhost_cdb_config *conf=(mod_vhost_cdb_config *) ap_get_module_config(r->server->module_config,&vhost_cdb_module);
int itr;

/* 
 * 1. Environment
 */
if( (conf->export_to_env!=NULL) && !apr_is_empty_array(conf->export_to_env))
	{
	for (itr=0; itr < conf->export_to_env->nelts; itr++) 
		{
		const char *key=((char **)conf->export_to_env->elts)[itr];
		const char *val=apr_table_get(db_conf,key);
		if(val)
			{
			/* ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"mod_vhost_cdb_export_vars: exporting %s to env",key); */
			apr_table_set(r->subprocess_env, key, val);
			}
		}
	}

/* 
 * 2. Notes
 */

if( (conf->export_to_notes!=NULL) && !apr_is_empty_array(conf->export_to_notes))
	{
	for (itr=0; itr < conf->export_to_notes->nelts; itr++) 
		{
		const char *key=((char **)conf->export_to_notes->elts)[itr];
		const char *val=apr_table_get(db_conf,key);
		if(val)
			{
			/* ap_log_rerror(APLOG_MARK,APLOG_DEBUG,0,r,"mod_vhost_cdb_export_vars: exporting %s=%s to notes",key,val); */
			apr_table_set(r->notes, key, val);
			}
		}
	}

return 1;
}

/* **************************************************************************** */
/* expand_path(r, apr_table_t, char *replacement)				*/
/* replaces %1 in DocumentRoot and ScriptAlias with replacement			*/
/* target strings are strlen(documentroot)+strlen(replacement)+1-strlen("%1")	*/
/* ****************************************************************************	*/

static int expand_path(request_rec *r, apr_table_t *conf, const char *replacement)
{
char *curr_dr=(char *)apr_table_get(conf,"DocumentRoot");
char *curr_sa=(char *)apr_table_get(conf,"ScriptAlias");

char *tag=strstr(curr_dr,"%1");
if(tag)
	{
	char *new=apr_pstrmemdup(r->pool, curr_dr, (tag-curr_dr));
	new=apr_pstrcat(r->pool,new,replacement,tag+2,NULL);
	apr_table_setn(conf,"DocumentRoot",new);
	}

tag=strstr(curr_sa,"%1");
if(tag)
	{
	char *new=apr_pstrmemdup(r->pool, curr_sa, (tag-curr_sa));
	new=apr_pstrcat(r->pool,new,replacement,tag+2,NULL);
	apr_table_setn(conf,"ScriptAlias",new);
	}

return 1;
}
/* ****************************************************************************	*/
/* get_config_from_db(struct cdb *db, const char *hostname)			*/
/* returns the entry from the db						*/
/* ****************************************************************************	*/

static char *get_config_from_db(request_rec *r,struct cdb *db, const char *hostname, mod_vhost_cdb_config *conf)
{
char *db_value=db_get(r, db, hostname);
if(db_value)
	{
	return db_value;
	}

/* 
 * try to match foo.domain.tld -> *.domain.tld
 */
char *dot_pos=(char *) hostname;

while( (dot_pos!=NULL) && (db_value==NULL) )
	{
	dot_pos=index(dot_pos,'.');
	if(dot_pos!=NULL)
		{
		++dot_pos;
		db_value=db_get(r, db, apr_pstrcat(r->pool,"*.",dot_pos,NULL));
		/* ap_log_error(APLOG_MARK, APLOG_DEBUG, 0,r->server, "get_config_from_db: checking for \"*.%s\"",dot_pos); */
		}
	}

/* matched */
if(dot_pos!=NULL)
	{
	apr_table_addn(r->notes,"VHostCDB_DomainPrefix",apr_pstrmemdup(r->pool,hostname,(int)(dot_pos-hostname-1)));
	return db_value;
	}

/*
 * did not match
 * so try to use the default_host if configured or return NULL
 * if using the default_host and it does not exist then return NULL
 */
if( (conf->defaulthost==NULL) || (strcmp(hostname,conf->defaulthost)==0) )
	{
	return NULL;
	}
	
/* ap_log_error(APLOG_MARK, APLOG_WARNING, 0,r->server, "get_config_from_db: using defaulthost \"%s\"",conf->defaulthost); */

db_value=db_get(r, db, conf->defaulthost);
if(!db_value)
	{
	ap_log_error(APLOG_MARK, APLOG_ERR, 0,r->server, "get_config_from_db: no entry for default host \"%s\"",conf->defaulthost);
	return NULL;
	}
return db_value;
}


/* ****************************************************************************	*/
/* mod_vhost_cdb_trans_uri							*/
/* sets the values from cdb							*/
/* ****************************************************************************	*/

static int mod_vhost_cdb_trans_uri(request_rec *r)
{
//ap_log_error(APLOG_MARK, APLOG_ERR, 0,r->server, "trans_uri: called");

mod_vhost_cdb_config *conf=(mod_vhost_cdb_config *) ap_get_module_config(r->server->module_config,&vhost_cdb_module);

if (conf->enable==0) return DECLINED;

/*
 * hostname:
 * 1) use the hostname from request_rec
 * 2) use the hostname from ap_get_server_name()
 * 3) use the defaulthost if nonempty or stop processing here
 */

const char *hostname=r->hostname;

if ( (hostname==NULL) || (strlen(hostname)==0) )
	{
	hostname=ap_get_server_name(r);
	if( (hostname==NULL) || (strlen(hostname)==0) )
		{
		/* ap_log_error(APLOG_MARK, APLOG_ERR, 0,r->server, "trans_uri: no hostname received"); */
		if(conf->defaulthost==NULL)
			{
			return DECLINED;
			}
		/* ap_log_error(APLOG_MARK, APLOG_WARNING, 0,r->server, "trans_uri: using defaulthost \"%s\"",conf->defaulthost); */
		hostname=conf->defaulthost;
		}
	};

/*
 * open connection to db
 */
struct cdb *db=apr_pcalloc(r->pool,sizeof(struct cdb));

int fd=open(conf->filename,O_RDONLY);
if(fd==-1)
	{
	ap_log_error(APLOG_MARK,APLOG_CRIT,0,r->server,"open failed: %s", strerror(errno));
	return HTTP_INTERNAL_SERVER_ERROR;
	}
if(cdb_init(db,fd)<0)
	{	
	ap_log_error(APLOG_MARK,APLOG_CRIT,0,r->server,"cdb_init failed: %s", strerror(errno));
	close(fd);
	return HTTP_INTERNAL_SERVER_ERROR;
	}

/*
 * now query db and close link afterwards
 */

char *db_value=get_config_from_db(r, db, hostname, conf);
cdb_free(db);
close(fd);

if(!db_value)
	{
	return DECLINED;
	}

/* we now have got a valid configuration string, 
 * so lets unpack it and apply the configuration to the vhost
 *
 * - If we got a DocumentRoot, translate uris and so on
 * - otherwise if there is a Redirect configured, redirect to the url
 * - if none of the above match, return DECLINED
 */

apr_table_t *db_conf=create_config_table(r, db_value);
if(!apr_table_get(db_conf,"DocumentRoot"))
	{
	char *redir_url;
	if((redir_url=(char *)apr_table_get(db_conf,"Redirect"))==NULL)
		{
		ap_log_error(APLOG_MARK, APLOG_WARN, 0,r->server, "trans_uri: no DocumentRoot or Redirect configured for \"%s\"",hostname);
		return DECLINED;
		}
	/* ap_log_error(APLOG_MARK, APLOG_DEBUG, 0,r->server, "trans_uri: redirect to \"%s\"",redir_url); */
	apr_table_set(r->headers_out,"Location",redir_url);
	return HTTP_MOVED_PERMANENTLY;
	}

/* if we came from a *.foobar.baz, then replace %1 in DocumentRoot 
 * and ScriptAlias with virtual_dir
 */
if(apr_table_get(r->notes,"VHostCDB_DomainPrefix"))
	{
	expand_path(r,db_conf,apr_table_get(r->notes,"VHostCDB_DomainPrefix"));
	}

/*
 * set_document_root returns 0 if uri does not start with /
 */
if(!mod_vhost_cdb_set_document_root(r,db_conf))
	{
	return DECLINED;
	}

mod_vhost_cdb_export_vars(r,db_conf);
	
/*
 * user and group have to be saved in request->notes, because
 * the suexec_hook is called at another stage
 */

char *user, *group;
if( (user=(char *)apr_table_get(db_conf,"UID")) && (group=(char *)apr_table_get(db_conf,"GID")) )
	{
	apr_table_set(r->notes, "VHostCDB_UID", user);
	apr_table_set(r->notes, "VHostCDB_GID", group);
	}

return OK;
};

/* ****************************************************************************	*/
/* mod_vhost_cdb_suexec								*/
/* suexec_id_handler for ap_hook_get_suexec_identity				*/
/* this code is borrowed from mod_vhost_ldap.c					*/
/* ****************************************************************************	*/

#ifdef HAVE_UNIX_SUEXEC
static ap_unix_identity_t *mod_vhost_cdb_suexec(const request_rec * r)
{
mod_vhost_cdb_config *conf=(mod_vhost_cdb_config *) ap_get_module_config(r->server->module_config,&vhost_cdb_module);

if(conf->enable != 1) return NULL;

const request_rec *select=r;

/*
 * when running php from a ScriptAlias+Action foo
 * we have to use the request_rec we came from
 */

if((select->prev) && (apr_table_get(select->prev->notes,"VHostCDB_UID")) )
	{
	select=select->prev;
	}


char *user=(char *) apr_table_get(select->notes,"VHostCDB_UID");
char *group=(char *) apr_table_get(select->notes,"VHostCDB_GID");

if(!user || !group)
	{
	ap_log_error(APLOG_MARK, APLOG_ERR, 0,r->server, "mod_vhost_cdb_suexec: notes->VHostCDB_UID/notes->VHostCDB_GID empty");
	return NULL;
	}

ap_unix_identity_t *ugid = NULL;

if((ugid=apr_palloc(r->pool, sizeof(ap_unix_identity_t))) == NULL) 
	{
	return NULL;
	}

uid_t uid = -1;
gid_t gid = -1;

uid = (uid_t)atoll(user);
gid = (gid_t)atoll(group);

ugid->uid = uid;
ugid->gid = gid;
ugid->userdir = 0;
  
return ugid;
}
#endif

/* ****************************************************************************	*/
/* mod_vhost_cdb_create_config							*/
/* hook zum Initialisieren der Konfiguration					*/
/* ****************************************************************************	*/

static void *mod_vhost_cdb_create_cfg(apr_pool_t *p, server_rec *s)
{
mod_vhost_cdb_config *conf=(mod_vhost_cdb_config *) apr_pcalloc(p,sizeof(mod_vhost_cdb_config));

conf->filename=apr_pstrcat(p,"/usr/local/etc/apache2/vhosts.cdb",NULL);
conf->enable=0;
conf->export_to_env=apr_array_make(p,0,sizeof(char *));
conf->export_to_notes=apr_array_make(p,0,sizeof(char *));
conf->defaulthost=NULL;

/* ap_log_error(APLOG_MARK, APLOG_DEBUG, 0,s, "mod_vhost_cdb: created vhost-config"); */

return (void *)conf;
};

/* ****************************************************************************	*/
/* mod_vhost_set_enable								*/
/* hook fuer die Konfigurationsdirektive VhostCDB_Enable			*/
/* ****************************************************************************	*/

static const char *mod_vhost_cdb_set_enable(cmd_parms *cmd, void *dummy, int enabled)
{
mod_vhost_cdb_config *conf = (mod_vhost_cdb_config *)ap_get_module_config(cmd->server->module_config, &vhost_cdb_module);
conf->enable = (enabled==1) ? 1 : 0;

return NULL;
}

/* ****************************************************************************	*/
/* mod_vhost_set_filename							*/
/* hook fuer die Konfigurationsdirektive VhostCDB_Filename			*/
/* ****************************************************************************	*/

static const char *mod_vhost_cdb_set_filename(cmd_parms *cmd, void *dummy, const char *filename)
{
mod_vhost_cdb_config *conf = (mod_vhost_cdb_config *)ap_get_module_config(cmd->server->module_config, &vhost_cdb_module);
conf->filename = apr_pstrdup(cmd->pool,filename);

/* ap_log_error(APLOG_MARK, APLOG_DEBUG,0,cmd->server,"mod_vhost_cdb_set_filename: conf->filename is now: %s",conf->filename); */

return NULL;
}

/* ****************************************************************************	*/
/* mod_vhost_set_defaulthost							*/
/* hook fuer die Konfigurationsdirektive VhostCDB_DefaultHost			*/
/* ****************************************************************************	*/

static const char *mod_vhost_cdb_set_defaulthost(cmd_parms *cmd, void *dummy, const char *hostname)
{
mod_vhost_cdb_config *conf = (mod_vhost_cdb_config *)ap_get_module_config(cmd->server->module_config, &vhost_cdb_module);
conf->defaulthost = apr_pstrdup(cmd->pool,hostname);

/* ap_log_error(APLOG_MARK, APLOG_DEBUG,0,cmd->server,"mod_vhost_cdb_set_filename: conf->filename is now: %s",conf->filename); */

return NULL;
}

/* ****************************************************************************	*/
/* mod_vhost_cdb_add_export_env							*/
/* hook fuer die Konfigurationsdirektive VhostCDB_ExportToEnv			*/
/* ****************************************************************************	*/

static const char *mod_vhost_cdb_add_export_env(cmd_parms *cmd, void *dummy, const char *key)
{
mod_vhost_cdb_config *conf = (mod_vhost_cdb_config *)ap_get_module_config(cmd->server->module_config, &vhost_cdb_module);

char **ptr=apr_array_push(conf->export_to_env);
*ptr=apr_pstrdup(cmd->pool,key);

/* ap_log_error(APLOG_MARK, APLOG_DEBUG,0,cmd->server,"mod_vhost_cdb_add_export_env: adding %s to list",key); */

return NULL;
}

/* ****************************************************************************	*/
/* mod_vhost_add_export_notes							*/
/* hook fuer die Konfigurationsdirektive VhostCDB_ExportToRequestNotes		*/
/* ****************************************************************************	*/

static const char *mod_vhost_cdb_add_export_notes(cmd_parms *cmd, void *dummy, const char *key)
{
mod_vhost_cdb_config *conf = (mod_vhost_cdb_config *)ap_get_module_config(cmd->server->module_config, &vhost_cdb_module);

char **ptr=apr_array_push(conf->export_to_notes);
*ptr=apr_pstrdup(cmd->pool,key);

/* ap_log_error(APLOG_MARK, APLOG_DEBUG,0,cmd->server,"mod_vhost_cdb_add_export_notes: adding %s to list",key); */

return NULL;
}


/* ****************************************************************************	*/
/* ****************************************************************************	*/
/* ****************************************************************************	*/


static void mod_vhost_cdb_register_hooks(apr_pool_t *p) 
{
#ifdef HAVE_UNIX_SUEXEC
ap_hook_get_suexec_identity(mod_vhost_cdb_suexec, NULL, NULL, APR_HOOK_MIDDLE);
#endif
ap_hook_translate_name(mod_vhost_cdb_trans_uri, NULL, NULL, APR_HOOK_MIDDLE);
};


static const command_rec mod_vhost_cdb_cmds[] = 
{
AP_INIT_FLAG("VhostCDB_Engine",(void *)mod_vhost_cdb_set_enable, NULL, RSRC_CONF, "set on or off to en/disable mod_vhost_cdb"),
AP_INIT_TAKE1("VhostCDB_File",(void *)mod_vhost_cdb_set_filename, NULL, RSRC_CONF, "set filename of cdb-file to use"),
AP_INIT_TAKE1("VhostCDB_DefaultHost",(void *)mod_vhost_cdb_set_defaulthost, NULL, RSRC_CONF, "set default host to use if no hostname was submitted or no entry was found in cdb"),
AP_INIT_ITERATE("VhostCDB_ExportToEnv",(void *)mod_vhost_cdb_add_export_env, NULL, RSRC_CONF, "set key(s) from cdb to export to environment"),
AP_INIT_ITERATE("VhostCDB_ExportToRequestNotes",(void *)mod_vhost_cdb_add_export_notes, NULL, RSRC_CONF, "set key(s) from cdb to export to request-notes"),
{ NULL }
};

module AP_MODULE_DECLARE_DATA vhost_cdb_module = {
	STANDARD20_MODULE_STUFF,
	NULL,
	NULL,
	mod_vhost_cdb_create_cfg, 
	NULL,
	mod_vhost_cdb_cmds,
	mod_vhost_cdb_register_hooks
};



